__version__ = "$Id: Grammar.py,v 0.2 1997/12/13 03:02:13 scott Exp scott $"

import time

#
# Production -- a Grammar is really just a list of productions.
# The expected structure is a symbol for the LHS and a list of 
# symbols or symbols for the RHS.
#
class Production:

    def __init__(self, LHS, RHS):
	self.LHS = LHS
	self.RHS = RHS
	self.func = None
	self.funcname = ""

    #
    # .setfunc(<callable>)  --used for the dynamic production 
    # of a parseengine directly from Grammar.mkengine(), instead of tables 
    # saved to a file.
    #
    def setfunc(self, func):
	self.func = func
	
    #
    # .setfuncname("") -- used by Grammar.writefile to produce
    # prodinfo table that.  .setfunc associates a function value
    # with the production for runtime, on the fly productions
    # of parsing engine from Grammar.
    #
    def setfuncname(self, name):
	self.funcname = name

    def __len__(self):
	return len(self.RHS)

    def __repr__(self):
	return self.LHS + " -> " + `self.RHS`

    def items(self):
	return range(len(self.RHS) + 1)

#
# Perhaps this should be called 'LR1Grammar' instead...
# Provides methods for producing the actiontable, the gototable, and the 
# prodinfo table.  Using these functions, it can produce a python source code file
# with these tables or a parsing engine.
#
class Grammar:

    EPS = "<EPS>"
    EOF = "<EOF>"
    DummyLA = -1

    def __init__(self, prods, toks=None):
	self.productions = prods[:]
	if toks:
	    self.tokens = toks
	else:
	    self.tokens = []
	self.terms = self.terminals()
	self.nonterms = self.nonterminals()
	self.lhsprods = {}
	for s in self.nonterms:
	    self.lhsprods[s] = self.LHSproductions(s)
	pi = 1                  
	self.epslhs = {}
	for p in self.productions:
	    if p.RHS == []:
		self.epslhs[p.LHS] = pi
	    pi = pi + 1
	self.lhsdereps = self.mklhsdereps()
	self.firstmap = self.mkfirstmap()
	self.verbose = 0

    #
    # the next three functions, expansions, showexp, and __repr__ all just print the grammar
    #
    def expansions(self, LHS):
	e = []
	for p in self.productions:
	    if p.LHS == LHS:
		e.append(p.RHS)
	return e

    def showexp(self, LHS):
	import string
	s = LHS + " -> "
	w = len(s)
	exps = self.expansions(LHS)[:]
	prodreprs = []
	for e in exps[:-1]:
	    for ei in range(len(e)):
		if type(e[ei]) == type(0):
		    if self.tokens:
			e[ei] = self.tokens[e[ei]]
		    else:
			e[ei] = str(e[ei])
	    
	    s = s  + (string.join(e, " ") + "\n        |" + " ")
	e = exps[-1]
	for ei in range(len(e)):
	    if type(e[ei]) == type(0):
		if self.tokens:
		    e[ei] = self.tokens[e[ei]]
		else:
		    e[ei] = str(e[ei])	    
	s = s + string.join(e, " ") + ";"
	return s


    def __repr__(self):
	import string
	s = ""
	for nt in self.nonterminals():
	    s = s + self.showexp(nt) + "\n"
	pi = 0
	lines = string.split(s, "\n")
	for l in lines[:-1]:
	    if self.productions[pi].funcname:
		lines[pi] = l + "%s(%s)" % (" " * (55 - len(l)), self.productions[pi].funcname)
	    pi = pi + 1
	return string.join(lines, "\n")

    #
    # this set of methods are used in initializing the data. In amny cases, each one 
    # depends on the existence of teh one prior
    #
    def terminals(self):
	t = []
	for p in self.productions:
	    for s in p.RHS:
		if s not in t and s != Grammar.EPS:
		    t.append(s)
	for nt in self.nonterminals():
	    if nt in t:
		t.remove(nt)
	t.sort()
	return t

    def nonterminals(self):
	nt = []
	for p in self.productions:
	    if p.LHS not in nt:
		nt.append(p.LHS)
	return nt


    def LHSproductions(self, lhs):
	"""
	returns the productions associated with a nonterminal
	"""
	res = []
	for p in self.productions:
	    if p.LHS == lhs:
		res.append(p)
	return res


    def mklhsdereps(self):
	"""
	determines the nonterminals that derive nothing (epsilon)
	"""
	pi = 1                  
	res = {}
	for p in self.productions:
	    if p.RHS == []:
		res[p.LHS] = pi
	    pi = pi + 1
	workingnonterms = []
	for nt in self.nonterms:
	    if not res.has_key(nt):
		workingnonterms.append(nt)
	while 1:
	    toremove = []
	    for nt in workingnonterms:
		if not res.has_key(nt):
		    for p in self.lhsprods[nt]:
			if len(p.RHS) == 1 and res.has_key(p.RHS[0]):
			    res[p.LHS] = res[p.RHS[0]]
			    toremove.append(nt)
			    break
	    if not toremove:
		break
	    for r in toremove:
		workingnonterms.remove(r)
	return res


    def mkfirstmap(self):
	"""
	returns a dictionary keyed by symbol whose values are the set
	of terminals that can start that symbol (d[nonterm] = [nonterm])
	"""
	res = {}
	for sym in self.terms + [Grammar.EPS, Grammar.EOF, Grammar.DummyLA]:
	    res[sym] = {sym: 1}
	while 1:
	    added = 0
	    for nt in self.nonterms:
		firsts = res.get(nt, {})
		for p in self.lhsprods[nt]:
		    if not p.RHS:
			if not firsts.has_key(Grammar.EPS):
			    added = 1
			    firsts[Grammar.EPS] = 1
			continue
		    i = 0
		    while i < len(p.RHS):
			f = res.get(p.RHS[i], {})
			for t in f.keys():
			    if not firsts.has_key(t):
				added = 1
				firsts[t] = 1
			if self.lhsdereps.has_key(p.RHS[i]):
			    i = i + 1
			else: break
		res[nt] = firsts
	    if not added:
		break
	for s in res.keys():
	    res[s] = res[s].keys()
	return res


    #
    # these function are used as the grammar produces the tables (or writes them
    # to a file)
    #
    def firstofstring(self, gs_list):
	tmpres = {}
	allhaveeps = 1
	for x in range(len(gs_list)):
	    tmp = self.firstmap[gs_list[x]]
	    for s in tmp:
		tmpres[s] = 1
	    if Grammar.EPS in tmp:
		del tmpres[Grammar.EPS]
	    else:
		allhaveeps = 0
		break
	if allhaveeps:
	    tmpres[Grammar.EPS] = 1
	return tmpres.keys()



    def augment(self):
	"""
	this function adds a production S' -> S to the grammar where S was
	the start symbol.
	"""
	lhss = map(lambda x: x.LHS, self.productions)
	newsym = self.productions[0].LHS
	while 1:
	    newsym = newsym + "'"
	    if newsym not in lhss:
		break
	self.productions.insert(0, Production(newsym, 
					      [self.productions[0].LHS]))


    #
    # follow is not used yet, but probably will be in determining error reporting/recovery
    #
    def follow(self):
	eof = Grammar.EOF
	follow = {}
	startsym = self.productions[0].LHS
	follow[startsym] = [eof]
	nts = self.nonterms
	for p in self.productions:
	    cutoff = range(len(p.RHS))
	    cutoff.reverse()
	    for c in cutoff[:-1]:  # all but the first of the RHS elements
		f = self.firstmap[p.RHS[c]]
		if Grammar.EPS in f:
		    f.remove(Grammar.EPS)
		if follow.has_key(p.RHS[c - 1]):
		    if p.RHS[c -1] in nts:
			follow[p.RHS[c -1]] = follow[p.RHS[c - 1]] + f[:]
		else:
		    if p.RHS[c -1] in nts:
			follow[p.RHS[c - 1]] = f[:]
 	for p in self.productions:
	    if not p.RHS: continue
 	    cutoff = range(len(p.RHS))
 	    cutoff.reverse()
	    if p.RHS[-1] in nts:
		if follow.has_key(p.LHS):
		    add = follow[p.LHS]
		else:
		    add = []

		if follow.has_key(p.RHS[-1]):
		    follow[p.RHS[-1]] = follow[p.RHS[-1]] + add
		else:
		    follow[p.RHS[-1]] = add
 	    for c in cutoff[:-1]:
 		f = self.firstmap[p.RHS[c]]
 		if Grammar.EPS in f:
 		    if follow.has_key(p.LHS):
 			add = follow[p.LHS]
 		    else:
 			add = []
 		    if follow.has_key(p.RHS[c-1]):
 			follow[p.RHS[c-1]] = follow[p.RHS[c-1]] + add
 		    elif add:
 			follow[p.RHS[c - 1]] = add
	for k in follow.keys():
	    d = {}
	    for i in follow[k]:
		d[i] = 1
	    follow[k] = d.keys()
	return follow


    def closure(self, items):
	res = items[:]
	todo = items[:]
	while 1:
	    more = []
	    for (prodind, rhsind), term in todo:
		if rhsind >= len(self.productions[prodind].RHS):
		    continue
		for p in self.lhsprods.get(self.productions[prodind].RHS[rhsind], []):
		    try:
			newpart = self.productions[prodind].RHS[rhsind + 1]
		    except IndexError:
			newpart = Grammar.EPS
		    stringofsyms = [newpart, term]
		    for t in self.firstofstring(stringofsyms):
			if ((self.productions.index(p), 0), t) not in res:
			    more.append(((self.productions.index(p), 0), t))
		    if term == Grammar.EOF and newpart == Grammar.EPS:
			if ((self.productions.index(p), 0), Grammar.EOF) not in res:
			    more.append(((self.productions.index(p), 0), Grammar.EOF))
	    if more:
		res = res + more
		todo = more
	    else:
		break
	return res


    def goto(self, items, sym):
	itemset = []
	for (prodind, rhsind), term in items:
	    try:
		if self.productions[prodind].RHS[rhsind] == sym and ((prodind, rhsind+1), term) not in itemset:
		    itemset.append( ((prodind, rhsind +1), term))
	    except IndexError:  
		pass
	return self.closure(itemset)




    def default_prodfunc(self):
	"""
	for mkengine, this will produce a default function for those unspecified
	"""
	return lambda *args: args[0]



    def prodinfotable(self):
	"""
	returns a list of three pieces of info for each production.
	The first is the lenght of the production, the second is the
	function(name) associated with the production and the third is
	is the index of the lhs in a list of nonterminals.
	"""
	res = []
	nonterms = self.nonterms
	for p in self.productions:
	    lhsind = nonterms.index(p.LHS)
	    func = p.func
	    if not func:
		func = self.default_prodfunc()
	    plen = len(p.RHS)
	    if p.RHS == [Grammar.EPS]:
		plen = 0
	    res.append((plen, func, lhsind))
	return res



class LALRGrammar(Grammar):

    def __init__(self, prods, toks=None):
	Grammar.__init__(self, prods, toks)
	self.LALRitems = []
	#
	# this is to help  mak epsilon productions work with kernel items and to compute goto
	# transitions from kernel
	#
	print "computing ntfirsts..."
	self.ntfirstmap = self.mkntfirstmap()  
	#
	# this is to help make shifts work with only kernel items
	#
	print "computing tfirsts..."
	self.tfirstmap = self.mktfirstmap()
	#
	# another thing to help epsilon productions
	#
	print "computing follows..."
	self.followmap = self.follow()
		


    def mkntfirstmap(self):
	"""
	computes all nonterms A, first of (strings n) such that some nonterminal B 
	derives [A, n] in zero or more steps of (rightmost) derivation. used to 
	help make epsilon productions quickly calculable. (B may == A)
	"""
	res = {}
	for p in self.productions:
	    if p.RHS and p.RHS[0] in self.nonterms:
		fos = self.firstofstring(p.RHS[1:])
		fos.sort()
		if not res.has_key(p.LHS):
		    res[p.LHS] = {}
		if not res[p.LHS].has_key(p.RHS[0]):
		    res[p.LHS][p.RHS[0]] = []
		for i in fos:
		    if i not in res[p.LHS].get(p.RHS[0], []):
			res[p.LHS][p.RHS[0]] = fos
		
	while 1:
	    foundmore = 0
	    reskeys = res.keys()
	    for nt in reskeys:
		rhsdict = res[nt]
		for rnt in rhsdict.keys():
		    if rnt in reskeys:
			d = res[rnt]
			for k in d.keys():
			    if not res[nt].has_key(k):
				fos = self.firstofstring(d[k]+ res[nt][rnt]) 
				foundmore = 1
 				fos.sort()
 				res[nt][k] = fos
 			    else:
 				fos = self.firstofstring(d[k] + res[nt][rnt])
 				fos.sort()
 				if fos != res[nt][k]:  # then res[nt][k] is contained in fos
 				    foundmore = 1
 				    res[nt][k] = fos
	    if not foundmore:
		break
	#
	# this part accounts for the fact that a nonterminal will
	# produce exactly itself in zero steps
	#
	for p in self.productions:
	    if res.has_key(p.LHS):
                res[p.LHS][p.LHS] = [Grammar.EPS]
	    else:
		res[p.LHS] = {p.LHS: [Grammar.EPS]}
	return res
			    

    def newmkntfirstmap(self):
	"""
	computes all nonterms A, first of (strings n) such that some nonterminal B 
	derives [A, n] in zero or more steps of (rightmost) derivation. used to 
	help make epsilon productions quickly calculable. (B may == A)
	"""
	res = {}
	pi = 0
	for p in self.productions:
	    if p.RHS and p.RHS[0] in self.nonterms:
		if not res.has_key(p.LHS):
		    res[p.LHS] = {}
		if not res[p.LHS].has_key(p.RHS[0]):
		    res[p.LHS][p.RHS[0]] = 1
	while 1:
	    foundmore = 0
	    reskeys = res.keys()
	    for nt in reskeys:
		rhsdict = res[nt]
		for rnt in rhsdict.keys():
		    if rnt in reskeys:
			d = res[rnt]
			for k in d.keys():
			    if not res[nt].has_key(k):
				foundmore = 1
 				res[nt][k] = 1
	    if not foundmore:
		break
	#
	# this part accounts for the fact that a nonterminal will
	# produce exactly itself in zero steps
	#
	for p in self.productions:
	    if res.has_key(p.LHS):
                res[p.LHS][p.LHS] = 1
	    else:
		res[p.LHS] = {p.LHS: 1}
	return res
	

	
    def mktfirstmap(self):
	"""
	for each nonterminal C, compute the set of all terminals a, such that
	C derives ax in zero or more steps of (rightmost) derivation where
	the last derivation is not an epsilon (empty) production.
	
	assumes .mkfirstntmap() has been run and has already produced self.ntfirstmap
	"""
	res = {}
	for p in self.productions:
	    if not res.has_key(p.LHS):
		res[p.LHS] = []
	    if p.RHS and p.RHS[0] in self.terms:
		res[p.LHS].append(p.RHS[0])
	while 1:
	    foundmore = 0
	    reskeys = res.keys()
	    for nt in self.ntfirstmap.keys():
		arrows = self.ntfirstmap[nt]
		for k in arrows.keys():
		    for t in res[k]:
			if t not in res[nt]:
			    foundmore = 1
			    res[nt].append(t)
	    if not foundmore:
		break
	return res

	
		
    def goto(self, itemset, sym):
	res = []
	for (pi, ri) in itemset:
	    if ri == len(self.productions[pi].RHS):
		continue
	    s = self.productions[pi].RHS[ri]
	    if s == sym:
		res.append((pi, ri+1))
            d = self.ntfirstmap.get(s, {})
	    for k in d.keys():
		for p in self.lhsprods[k]:
		    if p.RHS and p.RHS[0] == sym:
			i = self.productions.index(p)
			if (i, 1) not in res: res.append((i, 1))
	res.sort()
	return res

	    

    def lookaheads(self, itemset):
	setsofitems = kernels = self.kernelitems
	spontaneous = []
	propagates = {}
	gotomap = {}
	for (kpi, kri) in itemset:
	    C = self.closure([((kpi, kri), Grammar.DummyLA)])
	    for (cpi, cri), t in C:
		if (cri) == len(self.productions[cpi].RHS):
		    continue
		s = self.productions[cpi].RHS[cri]
		if gotomap.has_key(s):
		    newstate = gotomap[s]
		else:
		    newstate = setsofitems.index(self.goto(itemset, s))
		    gotomap[s] = newstate
		if t != Grammar.DummyLA:
		    spontaneous.append((newstate, (cpi, cri+1), t))
		else:
 		    if propagates.has_key((kpi, kri)):
			propagates[(kpi, kri)].append((newstate, (cpi, cri+1)))
		    else:
			propagates[(kpi, kri)]=[(newstate, (cpi, cri+1))]
	return spontaneous, propagates


    def kernelsoflalr1items(self):
	res = [[(0, 0)]]
	todo = [[(0, 0)]]
	while 1:
	    newtodo = []
	    for items in todo:
		for s in self.terms + self.nonterms + [Grammar.EOF]:
		    g = self.goto(items, s)
		    if g and g not in res:
			newtodo.append(g)
	    if not newtodo:
		break
	    else:
		if self.verbose:
		    print "found %d more kernels" % (len(newtodo))
		res = res + newtodo
		todo = newtodo
	res.sort()
	return res



    def initLALR1items(self):
	self.kernelitems = kernels = self.kernelsoflalr1items()
	props = {}
	la_table = []
	for x in range(len(kernels)):
	    la_table.append([])
	    for y in range(len(kernels[x])):
		la_table[x].append([])
	la_table[0][0] = [Grammar.EOF]
	if self.verbose:
	    print "initLALR1items, kernels done, calculating propagations and spontaneous lookaheads"
	state_i = 0
	for itemset in kernels:
	    if self.verbose:
		print ".",
            sp, pr = self.lookaheads(itemset)
	    for ns, (pi, ri), t in sp:
                inner = kernels[ns].index((pi, ri))
		la_table[ns][inner].append(t)
	    props[state_i] = pr
	    state_i = state_i + 1
	return la_table, props


    def LALR1items(self):
	la_table, props = self.initLALR1items()
	if self.verbose:
	    print "done init LALR1items"
	soi = self.kernelitems
	while 1:
	    added_la = 0
	    state_i = 0
	    for state in la_table:
		ii = 0
		for propterms in state:
		    if not propterms:
			ii = ii + 1
			continue
		    item = soi[state_i][ii]
		    ii = ii + 1
		    try:
			proplist = props[state_i][item]
		    except KeyError:
			continue
		    for pstate, pitem in proplist:
			inner = soi[pstate].index(pitem)
			for pt in propterms:
			    if pt not in la_table[pstate][inner]:
				added_la = 1
				la_table[pstate][inner].append(pt)
		state_i = state_i + 1
	    if not added_la:
		break
	#
	# this section just reorganizes the above data
	# to the state it's used in later...
	# 
	if self.verbose:
	    print "done with lalr1items, reorganizing the data"
	res = []
	state_i = 0
	for state in soi:
	    item_i = 0
	    inner = []
	    for item in state:
 		for term in la_table[state_i][item_i]:
 		    if (item, term) not in inner:
 			inner.append((item, term))
		item_i = item_i + 1
	    inner.sort()
	    res.append(inner)
	    state_i = state_i + 1
	self.LALRitems = res
	return res


    def deriveN(self, nt1, nt2):
	"""
	assuming nt1 -> nt2 <some string>, what is <some string>? such that
	we know it as 1) a set of terminals and 2) whether it contains
	Grammar.EPS
	"""
	


    def actiontable(self):
	items = self.LALRitems
	res = []
	state_i = 0
	terms = self.terms[:]
	terms.append(Grammar.EOF)
	errentry = ("", -1)
	for state in items:
	    list = [errentry] * len(terms)
	    res.append(list)
	    for (prodind, rhsind), term in state:
		if (rhsind ) == len(self.productions[prodind].RHS):
		    if prodind != 0:
			new = ("r", prodind)
			old = res[state_i][terms.index(term)]
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, terms.index(term)), old, "->", new
			res[state_i][terms.index(term)] = new
		    else:
			new = ("a", -1)
			old = res[state_i][terms.index(term)]
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, terms.index(term)), old, "->", new
			res[state_i][terms.index(term)] = new
		#
		# calculate reduction by epsilon productions 
		#
		elif self.productions[prodind].RHS[rhsind] in self.nonterms:
		    nt = self.productions[prodind].RHS[rhsind]
		    ntfirst = self.firstmap[nt]
		    ntfirsts = self.ntfirstmap.get(nt, {})
		    for k in ntfirsts.keys():
			if self.epslhs.get(k, ""):
			    reduceterms = self.followmap[k]
			    print `((prodind, rhsind), term)`, reduceterms
			    for r in reduceterms:
   				inner = terms.index(r)
   				old = res[state_i][inner]
   				new = ("r", self.epslhs[k])
  			        if old != errentry and old != new:
  				    print "Conflict[%d,%d]:" % (state_i, inner), old, "->", new
  				res[state_i][inner] = new
		    #
		    # calculate the shifts that occur but whose normal items aren't in the kernel
		    #
		    tfirsts = self.tfirstmap[nt]
		    for t in tfirsts:
			inner = terms.index(t)
			g = self.goto(self.kernelitems[state_i], t)
			old = res[state_i][inner]
			try:
			    news = self.kernelitems.index(g)
			except ValueError:
			    continue
			new = ("s", news)
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, inner), old, "->", new			    
			res[state_i][inner] = new
		#
		# compute the rest of the shifts that occur 'normally' in the kernel
		#
		else:
		    t = self.productions[prodind].RHS[rhsind]
		    inner = self.terms.index(t)
		    gt = self.goto(self.kernelitems[state_i], t)
		    if gt in self.kernelitems:
			news = self.kernelitems.index(gt)
			old = res[state_i][inner]
			new = ("s", news)
			if old != errentry and old != new:
			    print "Conflict[%d,%d]:" % (state_i, inner), old, "->", new			    
			res[state_i][inner] = new
	    state_i = state_i + 1
	return res



    def gototable(self):
	items = self.kernelitems
	res = []
	state_i = 0
	nonterms = self.nonterms
	err = None
	for state in items:
	    list = [err] * len(nonterms)
	    res.append(list)
	    nonterm_i = 0
	    for nt in nonterms:
		goto = self.goto(state, nt)
		if goto in items:
		    res[state_i][nonterm_i] = items.index(goto)
		nonterm_i = nonterm_i + 1
	    state_i = state_i + 1
	return res


    def mkengine(self, inbufchunksize=None, stackchunksize=None):
	""" 
	dynamically will produde a parse engine, just an experiment,
	don't try to use it for anything real.
	"""
	import PyLRengine
	self.augment()
	self.LALR1items()
	at = self.actiontable()
	gt = self.gototable()
	self.productions = self.productions[1:]  # unaugment
	pi = self.prodinfotable()
	if not inbufchunksize:
	    inbufchunksize = 50
	if not stackchunksize:
	    stackchunksize = 100
	e = PyLRengine.NewEngine(pi, at, gt, inbufchunksize, stackchunksize)
	return e


    def writefile(self, filename, parsername="MyParser", lexerinit = "PyLR.Lexer.Lexer()"):
	import time
	self.augment()
	print "About to start LALRitems at %d" % (time.time())
	self.LALR1items()
	print "done building LALRitems at %d" % (time.time())
	at = self.actiontable()
	print "done building actiontable at %d" % (time.time())
	gt = self.gototable()
	print "done building gototable at %d" % (time.time())
	self.productions = self.productions[1:]
	pi = self.prodinfotable()
	import parsertemplate
	template = parsertemplate.__doc__
	vals = {"parsername": parsername, "lexerinit": lexerinit}
	import time
	vals["date"] = time.ctime(time.time())
	vals["filename"] = filename
	if not hasattr(self, "extrasource"):
	    vals["extrasource"] = ""
	else:
	    vals["extrasource"] = self.extrasource
	vals["grammar"] = `self`
	actiontable_s = "[\n\t"
	for l in at:
	    actiontable_s = "%s%s,\n\t" % (actiontable_s, `l`)
	vals["actiontable"] = actiontable_s[:-3] + "\n]\n\n"
	gototable_s = "[\n\t"
	for l in gt:
	    gototable_s = "%s%s,\n\t" % (gototable_s, `l`)
	vals["gototable"] = gototable_s[:-3] + "\n]\n\n"
	pi_s = "[\n\t"
	pii = 0
	vals["symbols"] = `self.tokens + ["EOF"]`
	prod2func_s = "Production" + " " * 45 + "Method Name\n" 
	for l, f, e in pi:
	    pi_s = "%s(%d, '%s', %d),%s# %s\n\t" % (pi_s, 
						    l, 
						    self.productions[pii].funcname, 
						    e, 
						    " " * (18 - len(self.productions[pii].funcname)),
						    `self.productions[pii]` )
	    pii = pii + 1
	vals["prodinfo"] = pi_s + "]\n\n"
	fp = open(filename, "w")
	fp.write(template % vals)
	fp.close()

	

def _test():
    # the token interface will be defined in 
    # a wrapper module and in the lexer.
    # this is the format the information will be stored in
    #
    # first, define the tokens
    #
    (id, plus, times, lparen, rparen, eof) = range(6)
    #
    # then, define how you want them to appear
    # in the documentation of the output file
    #
    toks = ["id", "+", "*", "(", ")"]

    #
    # define the productions (LHS=left hand side, RHS=right hand side)
    #
    prods = map(lambda x: Production(x[0], x[1]), [("E", ["E", plus, "T"]),
						   ("E", ["T"]), 
						   ("T", ["T", times, "F"]), 
						   ("T", ["F"]),
						   ("F", [lparen, "E", rparen]),
						   ("F", [id])])
    global g
    g = LALRGrammar(prods, toks)
    #

    #
    # define functions for the parser to use
    #
    def fadd(*args):
	print "adding %d with %d" % (args[0], args[2])
	return args[0] + args[2]

    def fdummy(*args):
	print "calling fdummy with args %s" % `args`
	return args[0]

    def ftimes(*args):
	print "multiplying %d with %d" % (args[0], args[2])
	return args[0] * args[2]

    def fparens(*args):
	print "handling parens, returning whats in between"
	return args[1]

    
    #
    # register the functions
    #
    for i in range(len(g.productions)):
	if len(g.productions[i].RHS) == 1: 
	    g.setfunc(i, fdummy)
    g.setfunc(0, fadd)
    g.setfunc(2, ftimes)
    g.setfunc(4, fparens)
    #
    # produce the parser
    #
    print "generating parser"
    e = g.mkengine()
    print "parser generated"
    #
    # this is the input as would be returned by the Lexer 
    #
    input = [(lparen, "("), (id, 3), (plus, "+"),  
	     (lparen, "("), (id, 4), (times, "*"), (id, 2),  (rparen, ")"), 
	     (rparen, ")"), (times, "*"), 
	     (id, 2), (times, "*"), (id, 5) ]
    #
    # the parser keeps parsing until either an error occurs or an accept state occurs.  There
    # is no special EOF symbol for it to stop taking input yet, so for the moment you have to tag
    # a bunch of them on the end of the input :(
    #
    eofs = [(eof, "")] * 15
    input = input + eofs
    for token, val in input:
	e.parse(token, val)
    


    


   



    

